#include "Sim.hpp"
#include "Command.hpp"
#include "Cleanup.hpp"

using std::string;
using std::vector;

#define DEBUG 0

Sim::Sim() :
    m_energy(0),
    m_harmonics(false),
    m_model(),
    m_bots()
{
}

Sim::~Sim()
{
}

void Sim::init(uint32_t res)
{
    m_energy = 0;
    m_harmonics = false;
    m_model.resize(res);
    m_bots.resize(1);
    auto& bot = m_bots[0];
    bot.m_bid = 1;
    bot.m_pos = Coord();
    bot.m_seeds.resize(19);
    for (uint8_t seed = 2; seed <= 20; seed++)
    {
        bot.m_seeds[seed - 1] = seed;
    }
}

class FusionRec
{
public:
    Coord m_coordP;
    Coord m_coordS;
    size_t m_iBotP = (size_t)-1;
    size_t m_iBotS = (size_t)-1;
};

bool Sim::step(const vector<Command>& trace,
               size_t& pos,
               string* pMsg)
{
    size_t numBots = m_bots.size();
    size_t traceSize = trace.size();
    if (pos > traceSize || pos + m_bots.size() > traceSize)
    {
        if (pMsg) *pMsg = "Too few commands";
        return false;
    }

    //vector<Coord> volatileCoords;
    //vector<Bot> newBots;
    uint64_t newEnergy = m_energy;
    bool newHarmonics = m_harmonics;
    //vector<Coord> fills;
    //vector<FusionRec> fusionRecs;
    uint32_t res = m_model.getRes();

    Cleanup tempCleanup(
        [&](){
            volatileCoords.clear();
            newBots.clear();
            fills.clear();
            fusionRecs.clear();
        });

    {
        uint32_t factor = m_harmonics ? 30 : 3;
        newEnergy += factor * res * res * res;
    }

    newEnergy += 20 * numBots;

    for (size_t iBot = 0; iBot < numBots; iBot++)
    {
        auto& bot = m_bots[iBot];
        auto& command = trace[pos + iBot];
        Action action = command.m_action;
        switch (action)
        {
        case Action::Halt:
        {
#if DEBUG
            printf("Halt\n");
#endif
            if (numBots != 1)
            {
                if (pMsg) *pMsg = "Halt: too many bots";
                return false;
            }
            if (bot.m_pos != Coord())
            {
                if (pMsg) *pMsg = "Halt: bot not at origin";
                return false;
            }
            if (m_harmonics)
            {
                if (pMsg) *pMsg = "Halt: harmonics not low";
                return false;
            }
            volatileCoords.push_back(bot.m_pos);
            break;
        }
        case Action::Wait:
        {
#if DEBUG
            printf("Wait\n");
#endif
            volatileCoords.push_back(bot.m_pos);
            newBots.push_back(bot);
            break;
        }
        case Action::Flip:
        {
#if DEBUG
            printf("Flip\n");
#endif
            volatileCoords.push_back(bot.m_pos);
            newBots.push_back(bot);
            newHarmonics = !newHarmonics;
            break;
        }
        case Action::SMove:
        {
#if DEBUG
            printf("SMove (%" PRIi8 ", %" PRIi8 ", %" PRIi8 ")\n",
                   (int8_t)command.m_coord1.x,
                   (int8_t)command.m_coord1.y,
                   (int8_t)command.m_coord1.z);
#endif
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "SMove: destination out of range";
                return false;
            }
            Coord delta;
            uint32_t dist = 0;
            if (command.m_coord1.x != 0)
            {
                if (command.m_coord1.x <= 15)
                {
                    delta = Coord(1, 0, 0);
                    dist = command.m_coord1.x;
                }
                else
                {
                    delta = Coord(255, 0, 0);
                    dist = 256 - command.m_coord1.x;
                }
            }
            else if (command.m_coord1.y != 0)
            {
                if (command.m_coord1.y <= 15)
                {
                    delta = Coord(0, 1, 0);
                    dist = command.m_coord1.y;
                }
                else
                {
                    delta = Coord(0, 255, 0);
                    dist = 256 - command.m_coord1.y;
                }
            }
            else if (command.m_coord1.z != 0)
            {
                if (command.m_coord1.z <= 15)
                {
                    delta = Coord(0, 0, 1);
                    dist = command.m_coord1.z;
                }
                else
                {
                    delta = Coord(0, 0, 255);
                    dist = 256 - command.m_coord1.z;
                }
            }
            newEnergy += 2 * dist;
            Coord coord = bot.m_pos;
            volatileCoords.push_back(coord);
            for (uint32_t i = 0; i < dist; i++)
            {
                coord += delta;
                if (m_model.get(coord.x, coord.y, coord.z))
                {
                    if (pMsg) *pMsg = "SMove: voxel is full";
                    return false;
                }
                volatileCoords.push_back(coord);
            }
            newBots.push_back(bot);
            auto& newBot = newBots.back();
            newBot.m_pos = coord;
            break;
        }
        case Action::LMove:
        {
#if DEBUG
            printf("LMove (%" PRIi8 ", %" PRIi8 ", %" PRIi8 ") (%" PRIi8 ", %" PRIi8 ", %" PRIi8 ")\n",
                   (int8_t)command.m_coord1.x,
                   (int8_t)command.m_coord1.y,
                   (int8_t)command.m_coord1.z,
                   (int8_t)command.m_coord2.x,
                   (int8_t)command.m_coord2.y,
                   (int8_t)command.m_coord2.z);
#endif
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "LMove: bend out of range";
                return false;
            }
            newX += (int8_t)command.m_coord2.x;
            newY += (int8_t)command.m_coord2.y;
            newZ += (int8_t)command.m_coord2.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "LMove: destination out of range";
                return false;
            }
            Coord coord = bot.m_pos;
            volatileCoords.push_back(coord);
            for (uint32_t leg = 0; leg < 2; leg++)
            {
                Coord delta;
                uint32_t dist = 0;
                Coord commandCoord =
                    leg == 0 ? command.m_coord1 : command.m_coord2;
                if (commandCoord.x != 0)
                {
                    if (commandCoord.x <= 15)
                    {
                        delta = Coord(1, 0, 0);
                        dist = commandCoord.x;
                    }
                    else
                    {
                        delta = Coord(255, 0, 0);
                        dist = 256 - commandCoord.x;
                    }
                }
                else if (commandCoord.y != 0)
                {
                    if (commandCoord.y <= 15)
                    {
                        delta = Coord(0, 1, 0);
                        dist = commandCoord.y;
                    }
                    else
                    {
                        delta = Coord(0, 255, 0);
                        dist = 256 - commandCoord.y;
                    }
                }
                else if (commandCoord.z != 0)
                {
                    if (commandCoord.z <= 15)
                    {
                        delta = Coord(0, 0, 1);
                        dist = commandCoord.z;
                    }
                    else
                    {
                        delta = Coord(0, 0, 255);
                        dist = 256 - commandCoord.z;
                    }
                }
                newEnergy += 2 * dist;
                for (uint32_t i = 0; i < dist; i++)
                {
                    coord += delta;
                    if (m_model.get(coord.x, coord.y, coord.z))
                    {
                        if (pMsg) *pMsg = "LMove: voxel is full";
                        return false;
                    }
                    volatileCoords.push_back(coord);
                }
            }
            newEnergy += 2 * 2;
            newBots.push_back(bot);
            auto& newBot = newBots.back();
            newBot.m_pos = coord;
            break;
        }
        case Action::FusionP:
        {
#if DEBUG
            printf("FusionP (%" PRIi8 ", %" PRIi8 ", %" PRIi8 ")\n",
                   (int8_t)command.m_coord1.x,
                   (int8_t)command.m_coord1.y,
                   (int8_t)command.m_coord1.z);
#endif
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "FusionP: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;

            size_t iRec = 0;
            while (iRec < fusionRecs.size())
            {
                if (fusionRecs[iRec].m_coordP == bot.m_pos &&
                    fusionRecs[iRec].m_coordS == coord)
                {
                    break;
                }
            }
            if (iRec == fusionRecs.size())
            {
                fusionRecs.emplace_back();
                auto& rec = fusionRecs.back();
                rec.m_coordP = bot.m_pos;
                rec.m_coordS = coord;
            }
            fusionRecs[iRec].m_iBotP = iBot;

            volatileCoords.push_back(bot.m_pos);
            volatileCoords.push_back(coord);
            newEnergy -= 24;
            break;
        }
        case Action::FusionS:
        {
#if DEBUG
            printf("FusionS (%" PRIi8 ", %" PRIi8 ", %" PRIi8 ")\n",
                   (int8_t)command.m_coord1.x,
                   (int8_t)command.m_coord1.y,
                   (int8_t)command.m_coord1.z);
#endif
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "FusionS: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;

            size_t iRec = 0;
            while (iRec < fusionRecs.size())
            {
                if (fusionRecs[iRec].m_coordP == coord &&
                    fusionRecs[iRec].m_coordS == bot.m_pos)
                {
                    break;
                }
            }
            if (iRec == fusionRecs.size())
            {
                fusionRecs.emplace_back();
                auto& rec = fusionRecs.back();
                rec.m_coordP = coord;
                rec.m_coordS = bot.m_pos;
            }
            fusionRecs[iRec].m_iBotS = iBot;

            break;
        }
        case Action::Fission:
        {
#if DEBUG
            printf("Fission (%" PRIi8 ", %" PRIi8 ", %" PRIi8 ") %" PRIu8 "\n",
                   (int8_t)command.m_coord1.x,
                   (int8_t)command.m_coord1.y,
                   (int8_t)command.m_coord1.z,
                   command.m_m);
#endif
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "Fission: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;
            if (m_model.get(coord.x, coord.y, coord.z))
            {
                if (pMsg) *pMsg = "Fission: voxel is full";
                return false;
            }
            uint32_t m = command.m_m;
            if ((uint32_t)bot.m_seeds.size() < m + 1)
            {
                if (pMsg) *pMsg = "Fission: too few seeds";
                return false;
            }
            volatileCoords.push_back(bot.m_pos);
            volatileCoords.push_back(coord);
            newBots.push_back(bot);
            auto& newBot = newBots.back();
            newBot.m_seeds.erase(newBot.m_seeds.begin(),
                                 newBot.m_seeds.begin() + (m + 1));
            newBots.push_back(bot);
            auto& newBot2 = newBots.back();
            newBot2.m_bid = bot.m_seeds[0];
            newBot2.m_pos = coord;
            newBot2.m_seeds.erase(newBot2.m_seeds.begin() + (m + 1),
                                  newBot2.m_seeds.end());
            newEnergy += 24;
            break;
        }
        case Action::Fill:
        {
#if DEBUG
            printf("Fill (%" PRIi8 ", %" PRIi8 ", %" PRIi8 ")\n",
                   (int8_t)command.m_coord1.x,
                   (int8_t)command.m_coord1.y,
                   (int8_t)command.m_coord1.z);
#endif
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "Fill: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;
            volatileCoords.push_back(bot.m_pos);
            volatileCoords.push_back(coord);
            fills.push_back(coord);
            newBots.push_back(bot);
            break;
        }
        }
    }

    for (auto& rec : fusionRecs)
    {
        if (rec.m_iBotP == (size_t)-1 ||
            rec.m_iBotS == (size_t)-1)
        {
            if (pMsg) *pMsg = "Fusion mismatch";
            return false;
        }
        auto& botP = m_bots[rec.m_iBotP];
        auto& botS = m_bots[rec.m_iBotS];
        newBots.push_back(botP);
        auto& newBot = newBots.back();
        newBot.m_seeds.push_back(botS.m_bid);
        newBot.m_seeds.insert(newBot.m_seeds.end(),
                              botS.m_seeds.begin(),
                              botS.m_seeds.end());
        std::sort(newBot.m_seeds.begin(), newBot.m_seeds.end());
    }

    std::sort(volatileCoords.begin(), volatileCoords.end());
    auto adjIt = std::adjacent_find(volatileCoords.begin(), volatileCoords.end());
    if (adjIt != volatileCoords.end())
    {
        if (pMsg) *pMsg = "Coordinates interfere";
        return false;
    }

    newEnergy += 6 * fills.size();
    if (m_harmonics || newHarmonics)
    {
        for (auto& fill : fills)
        {
            if (!m_model.get(fill.x, fill.y, fill.z))
            {
                m_model.set(fill.x, fill.y, fill.z, 1);
                newEnergy += 6;
            }
        }
    }
    else
    {
        bool didSomething = true;
        bool skippedSomething = true;
        while (didSomething && skippedSomething)
        {
            didSomething = false;
            skippedSomething = false;
            for (auto& fill : fills)
            {
                if (!m_model.get(fill.x, fill.y, fill.z))
                {
                    uint8_t ok =
                        (fill.y == 0) ||
                        (fill.y > 0 &&
                         m_model.get(fill.x, fill.y - 1, fill.z)) ||
                        ((uint32_t)fill.y + 1 < res &&
                         m_model.get(fill.x, fill.y + 1, fill.z)) ||
                        (fill.x > 0 &&
                         m_model.get(fill.x - 1, fill.y, fill.z)) ||
                        ((uint32_t)fill.x + 1 < res &&
                         m_model.get(fill.x + 1, fill.y, fill.z)) ||
                        (fill.z > 0 &&
                         m_model.get(fill.x, fill.y, fill.z - 1)) ||
                        ((uint32_t)fill.z + 1 < res &&
                         m_model.get(fill.x, fill.y, fill.z + 1));
                    if (ok)
                    {
                        m_model.set(fill.x, fill.y, fill.z, 1);
                        newEnergy += 6;
                        didSomething = true;
                    }
                    else
                    {
                        skippedSomething = true;
                    }
                }
            }
        }
        if (skippedSomething)
        {
            if (pMsg) *pMsg = "Not grounded";
            return false;
        }
    }

    if (m_harmonics && !newHarmonics)
    {
        Model tmpModel(m_model);

        vector<Coord> q;
        for (uint32_t x = 0; x < res; x++)
        {
            for (uint32_t z = 0; z < res; z++)
            {
                if (tmpModel.get(x, 0, z))
                {
                    tmpModel.set(x, 0, z, 0);
                    q.push_back(Coord(x, 0, z));
                }
            }
        }

        vector<Coord> q2;
        while (!q.empty())
        {
            for (auto& coord : q)
            {
                uint32_t x = coord.x;
                uint32_t y = coord.y;
                uint32_t z = coord.z;
                if (x > 0 && tmpModel.get(x - 1, y, z))
                {
                    tmpModel.set(x - 1, y, z, 0);
                    q2.push_back(Coord(x - 1, y, z));
                }
                if (x + 1 < res && tmpModel.get(x + 1, y, z))
                {
                    tmpModel.set(x + 1, y, z, 0);
                    q2.push_back(Coord(x + 1, y, z));
                }
                if (y > 0 && tmpModel.get(x, y - 1, z))
                {
                    tmpModel.set(x, y - 1, z, 0);
                    q2.push_back(Coord(x, y - 1, z));
                }
                if (y + 1 < res && tmpModel.get(x, y + 1, z))
                {
                    tmpModel.set(x, y + 1, z, 0);
                    q2.push_back(Coord(x, y + 1, z));
                }
                if (z > 0 && tmpModel.get(x, y, z - 1))
                {
                    tmpModel.set(x, y, z - 1, 0);
                    q2.push_back(Coord(x, y, z - 1));
                }
                if (z + 1 < res && tmpModel.get(x, y, z + 1))
                {
                    tmpModel.set(x, y, z + 1, 0);
                    q2.push_back(Coord(x, y, z + 1));
                }
            }
            q.clear();
            q.swap(q2);
        }

        for (uint32_t x = 0; x < res; x++)
        {
            for (uint32_t y = 0; y < res; y++)
            {
                for (uint32_t z = 0; z < res; z++)
                {
                    if (tmpModel.get(x, y, z))
                    {
                        if (pMsg) *pMsg = "Not grounded when switching off harmonics";
                        return false;
                    }
                }
            }
        }
    }

    std::sort(newBots.begin(), newBots.end(),
              [](const Bot& a, const Bot& b){return a.m_bid < b.m_bid;});

    m_bots.swap(newBots);
    m_energy = newEnergy;
    m_harmonics = newHarmonics;

    pos += numBots;
    return true;
}
