#include "Model.hpp"
#include "Cleanup.hpp"

using std::string;
using std::unique_ptr;

Model::Model(const Model& other) :
    m_res(other.m_res),
    m_data(nullptr)
{
    if (m_res > 0)
    {
        uint32_t size = (m_res * m_res * m_res + 7) / 8;
        m_data.reset(new uint8_t[size]);
        for (size_t i = 0; i < size; i++)
        {
            m_data[i] = other.m_data[i];
        }
    }
}

Model::Model(Model&& other) :
    m_res(other.m_res),
    m_data(std::move(other.m_data))
{
    other.m_res = 0;
}

Model& Model::operator=(const Model& other)
{
    if (this != &other)
    {
        m_res = other.m_res;
        if (m_res > 0)
        {
            uint32_t size = (m_res * m_res * m_res + 7) / 8;
            m_data.reset(new uint8_t[size]);
            for (size_t i = 0; i < size; i++)
            {
                m_data[i] = other.m_data[i];
            }
        }
        else
        {
            m_data.reset();
        }
    }
    return *this;
}

Model& Model::operator=(Model&& other)
{
    if (this != &other)
    {
        m_res = other.m_res;
        m_data = std::move(other.m_data);
        other.m_res = 0;
    }
    return *this;
}

void Model::resize(uint32_t res)
{
    m_res = res;
    if (res > 0)
    {
        uint32_t size = (res * res * res + 7) / 8;
        m_data.reset(new uint8_t[size]());
    }
    else
    {
        m_data.reset();
    }
}

bool Model::load(const string& fileName,
                 string* pMsg)
{
    FILE* f = fopen(fileName.c_str(), "rb");
    if (!f)
    {
        if (pMsg) *pMsg = "Error opening model file";
        return false;
    }
    Cleanup fileCleanup([&](){fclose(f);});

    fseek(f, 0, SEEK_END);
    size_t size = (size_t)ftell(f);
    fseek(f, 0, SEEK_SET);

    if (size < 1)
    {
        if (pMsg) *pMsg = "Model file too small for resolution byte";
        return false;
    }

    uint8_t resByte = 0;
    if (fread(&resByte, 1, 1, f) != 1)
    {
        if (pMsg) *pMsg = "Error reading resolution byte";
        return false;
    }

    uint32_t res = resByte;
    if (res < 1 || res > 250)
    {
        if (pMsg) *pMsg = "Invalid resolution byte";
        return false;
    }

    size_t dataSize = (res * res * res + 7) / 8;
    if (size < 1 + dataSize)
    {
        if (pMsg) *pMsg = "Model file too small for data";
        return false;
    }
    if (size > 1 + dataSize)
    {
        if (pMsg) *pMsg = "Model file too large for data";
        return false;
    }

    unique_ptr<uint8_t[]> data(new uint8_t[dataSize]);
    if (fread(data.get(), 1, dataSize, f) != dataSize)
    {
        if (pMsg) *pMsg = "Error reading model data";
        return false;
    }

    fileCleanup.cleanup();

    for (size_t pos = res * res * res; pos < dataSize * 8; pos++)
    {
        uint32_t bytePos = pos >> 3;
        uint32_t bitPos = pos & 7;
        data[bytePos] &= ~((uint8_t)1 << bitPos);
    }

    m_res = res;
    m_data = std::move(data);

    return true;
}

bool Model::empty() const
{
    size_t dataSize = (m_res * m_res * m_res + 7) / 8;
    for (size_t i = 0; i < dataSize; i++)
    {
        if (m_data[i] != 0)
        {
            return false;
        }
    }

    return true;
}

bool Model::operator==(const Model& other) const
{
    if (m_res != other.m_res)
    {
        return false;
    }

    if (m_res < 1)
    {
        return true;
    }

    size_t dataSize = (m_res * m_res * m_res + 7) / 8;
    return std::equal(m_data.get(), m_data.get() + dataSize,
                      other.m_data.get());
}
