#include "Optimize.hpp"
#include "Model.hpp"
#include "Command.hpp"
#include "Sim.hpp"

using std::string;
using std::vector;

#define DEBUG 0

namespace
{
    class Step
    {
    public:
        size_t m_start;
        size_t m_end;
        size_t m_count;
        size_t m_avail;
        size_t m_numBots;
        bool m_needHarmonics;
    };
}

bool optimizeHarmonics(bool full,
                       const Model& sourceModel,
                       const vector<Command>& trace,
                       vector<Command>* pOptimizedTrace,
                       string* pMsg)
{
    vector<Step> steps;

    uint32_t res = sourceModel.getRes();
    Model searchModel(res);
    vector<Coord> q;
    vector<Coord> q2;

    Sim sim(full);
    sim.init(sourceModel);
    size_t pos = 0;
    while (pos < trace.size())
    {
        if (sim.m_bots.empty())
        {
            if (pMsg) *pMsg = "Halted before end of trace";
            return false;
        }
        steps.emplace_back();
        auto& step = steps.back();
        step.m_start = pos;
        if (!sim.step(trace, pos, pMsg))
        {
            return false;
        }
        step.m_end = pos;
        step.m_count = step.m_end - step.m_start;
        step.m_avail = (size_t)-1;
        for (size_t i = step.m_start; i < step.m_end; i++)
        {
            if (trace[i].m_action == Action::Wait ||
                trace[i].m_action == Action::Flip)
            {
                step.m_avail = i;
                break;
            }
        }
        step.m_numBots = sim.m_bots.size();

        bool needSearch = false;
        for (size_t i = step.m_start; i < step.m_end; i++)
        {
            Action action = trace[i].m_action;
            if (action == Action::Fill ||
                action == Action::Void ||
                action == Action::GFill ||
                action == Action::GVoid)
            {
                needSearch = true;
                break;
            }
        }

        if (!needSearch)
        {
            if (steps.size() >= 2)
            {
                step.m_needHarmonics = steps[steps.size() - 2].m_needHarmonics;
            }
            else
            {
                step.m_needHarmonics = false;
            }
        }
        else
        {
            searchModel = sim.m_model;

            for (uint32_t x = 0; x < res; x++)
            {
                for (uint32_t z = 0; z < res; z++)
                {
                    if (searchModel.get(x, 0, z))
                    {
                        searchModel.set(x, 0, z, 0);
                        q.push_back(Coord(x, 0, z));
                    }
                }
            }

            while (!q.empty())
            {
                for (auto& coord : q)
                {
                    uint32_t x = coord.x;
                    uint32_t y = coord.y;
                    uint32_t z = coord.z;
                    if (x > 0 && searchModel.get(x - 1, y, z))
                    {
                        searchModel.set(x - 1, y, z, 0);
                        q2.push_back(Coord(x - 1, y, z));
                    }
                    if (x + 1 < res && searchModel.get(x + 1, y, z))
                    {
                        searchModel.set(x + 1, y, z, 0);
                        q2.push_back(Coord(x + 1, y, z));
                    }
                    if (y > 0 && searchModel.get(x, y - 1, z))
                    {
                        searchModel.set(x, y - 1, z, 0);
                        q2.push_back(Coord(x, y - 1, z));
                    }
                    if (y + 1 < res && searchModel.get(x, y + 1, z))
                    {
                        searchModel.set(x, y + 1, z, 0);
                        q2.push_back(Coord(x, y + 1, z));
                    }
                    if (z > 0 && searchModel.get(x, y, z - 1))
                    {
                        searchModel.set(x, y, z - 1, 0);
                        q2.push_back(Coord(x, y, z - 1));
                    }
                    if (z + 1 < res && searchModel.get(x, y, z + 1))
                    {
                        searchModel.set(x, y, z + 1, 0);
                        q2.push_back(Coord(x, y, z + 1));
                    }
                }
                q.clear();
                q.swap(q2);
            }

            step.m_needHarmonics = !searchModel.empty();
        }
    }

    if (!sim.m_bots.empty())
    {
        if (pMsg) *pMsg = "Trace did not halt";
        return false;
    }

    auto& optimizedTrace = *pOptimizedTrace;
    bool prevHarmonics = false;
    for (size_t iStep = 0; iStep < steps.size(); iStep++)
    {
        auto& step = steps[iStep];

        bool extraFlip = false;

        if (step.m_needHarmonics &&
            !prevHarmonics &&
            step.m_avail == (size_t)-1)
        {
            extraFlip = true;
        }

        if (!step.m_needHarmonics &&
            prevHarmonics &&
            step.m_avail == (size_t)-1)
        {
            if (iStep > 0 &&
                steps[iStep - 1].m_needHarmonics)
            {
            }
            else if (iStep + 1 < steps.size() &&
                     steps[iStep + 1].m_needHarmonics &&
                     steps[iStep + 1].m_avail == (size_t)-1)
            {
            }
            else
            {
                extraFlip = true;
            }
        }

        if (extraFlip)
        {
#if DEBUG
            printf("\n");
#endif
            size_t numBots = (iStep == 0) ? 1 : steps[iStep - 1].m_numBots;
            optimizedTrace.push_back(Command(Action::Flip));
#if DEBUG
            printf("%s\n", optimizedTrace.back().toString().c_str());
#endif
            for (size_t i = 1; i < numBots; i++)
            {
                optimizedTrace.push_back(Command(Action::Wait));
#if DEBUG
                printf("%s\n", optimizedTrace.back().toString().c_str());
#endif
            }
            prevHarmonics = !prevHarmonics;
        }

        bool selfFlip = false;
        if (step.m_avail != (size_t)-1)
        {
            if (step.m_needHarmonics &&
                !prevHarmonics)
            {
                selfFlip = true;
            }

            if (!step.m_needHarmonics)
            {
                if (iStep + 1 < steps.size() &&
                    steps[iStep + 1].m_needHarmonics &&
                    steps[iStep + 1].m_avail == (size_t)-1)
                {
                    selfFlip = !prevHarmonics;
                }
                else
                {
                    selfFlip = prevHarmonics;
                }
            }
        }

#if DEBUG
        printf("\n");
#endif
        for (size_t i = step.m_start; i < step.m_end; i++)
        {
            if (step.m_avail != (size_t)-1 &&
                i == step.m_avail)
            {
                if (selfFlip)
                {
                    optimizedTrace.push_back(Command(Action::Flip));
                }
                else
                {
                    optimizedTrace.push_back(Command(Action::Wait));
                }
            }
            else if (trace[i].m_action == Action::Flip)
            {
                optimizedTrace.push_back(Command(Action::Wait));
            }
            else
            {
                optimizedTrace.push_back(trace[i]);
            }
#if DEBUG
            printf("%s\n", optimizedTrace.back().toString().c_str());
#endif
        }

        if (selfFlip)
        {
            prevHarmonics = !prevHarmonics;
        }
    }

    return true;
}
