#include "Sim.hpp"
#include "Command.hpp"
#include "Cleanup.hpp"

using std::string;
using std::vector;
using std::pair;

#define DEBUG 0

Sim::Sim(bool full) :
    m_full(full),
    m_energy(0),
    m_harmonics(false),
    m_model(),
    m_bots()
{
}

Sim::~Sim()
{
}

void Sim::init(uint32_t res)
{
    m_model.resize(res);
    initImpl();
}

void Sim::init(const Model& model)
{
    m_model = model;
    initImpl();
}

void Sim::init(Model&& model)
{
    m_model = std::move(model);
    initImpl();
}

void Sim::initImpl()
{
    m_energy = 0;
    m_harmonics = false;
    m_bots.resize(1);
    auto& bot = m_bots[0];
    bot.m_bid = 1;
    bot.m_pos = Coord();
    uint32_t numSeeds = m_full ? 40 : 20;
    bot.m_seeds.resize(numSeeds - 1);
    for (uint8_t seed = 2; seed <= numSeeds; seed++)
    {
        bot.m_seeds[seed - 2] = seed;
    }
}

class FusionRec
{
public:
    Coord m_coordP;
    Coord m_coordS;
    size_t m_iBotP = (size_t)-1;
    size_t m_iBotS = (size_t)-1;
};

class GroupRec
{
public:
    Coord m_coordLow;
    Coord m_coordHigh;
    Action m_action;
    size_t m_iBots[2][2][2] =
    {{{(size_t)-1, (size_t)-1},
      {(size_t)-1, (size_t)-1}},
     {{(size_t)-1, (size_t)-1},
      {(size_t)-1, (size_t)-1}}};
};

bool Sim::step(const vector<Command>& trace,
               size_t& pos,
               string* pMsg)
{
    size_t numBots = m_bots.size();
    size_t traceSize = trace.size();
    if (pos > traceSize || pos + m_bots.size() > traceSize)
    {
        if (pMsg) *pMsg = "Too few commands";
        return false;
    }

#if DEBUG
    for (size_t iBot = 0; iBot < numBots; iBot++)
    {
        auto& bot = m_bots[iBot];
        printf("%2" PRIu8 "  %s ", bot.m_bid, bot.m_pos.toString().c_str());
        for (uint8_t seed : bot.m_seeds)
        {
            printf(" %2" PRIu8 "", seed);
        }
        printf("\n");
    }
#endif

    //vector<Coord> volatileCoords;
    //vector<Bot> newBots;
    uint64_t newEnergy = m_energy;
    bool newHarmonics = m_harmonics;
    //vector<pair<Coord, Coord>> fills;
    //vector<pair<Coord, Coord>> voids;
    //vector<FusionRec> fusionRecs;
    //vector<GroupRec> groupRecs;
    uint32_t res = m_model.getRes();

    Cleanup tempCleanup(
        [&](){
            volatileCoords.clear();
            newBots.clear();
            fills.clear();
            voids.clear();
            fusionRecs.clear();
            groupRecs.clear();
        });

    {
        uint32_t factor = m_harmonics ? 30 : 3;
        newEnergy += factor * res * res * res;
    }

    newEnergy += 20 * numBots;

    for (size_t iBot = 0; iBot < numBots; iBot++)
    {
        auto& bot = m_bots[iBot];
        auto& command = trace[pos + iBot];
#if DEBUG
        printf("%s\n", command.toString().c_str());
#endif
        Action action = command.m_action;
        switch (action)
        {
        case Action::Halt:
        {
            if (numBots != 1)
            {
                if (pMsg) *pMsg = "Halt: too many bots";
                return false;
            }
            if (bot.m_pos != Coord())
            {
                if (pMsg) *pMsg = "Halt: bot not at origin";
                return false;
            }
            if (m_harmonics)
            {
                if (pMsg) *pMsg = "Halt: harmonics not low";
                return false;
            }
            volatileCoords.push_back(bot.m_pos);
            break;
        }
        case Action::Wait:
        {
            volatileCoords.push_back(bot.m_pos);
            newBots.push_back(bot);
            break;
        }
        case Action::Flip:
        {
            volatileCoords.push_back(bot.m_pos);
            newBots.push_back(bot);
            newHarmonics = !newHarmonics;
            break;
        }
        case Action::SMove:
        {
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "SMove: destination out of range";
                return false;
            }
            Coord delta;
            uint32_t dist = 0;
            if (command.m_coord1.x != 0)
            {
                if (command.m_coord1.x <= 15)
                {
                    delta = Coord(1, 0, 0);
                    dist = command.m_coord1.x;
                }
                else
                {
                    delta = Coord(255, 0, 0);
                    dist = 256 - command.m_coord1.x;
                }
            }
            else if (command.m_coord1.y != 0)
            {
                if (command.m_coord1.y <= 15)
                {
                    delta = Coord(0, 1, 0);
                    dist = command.m_coord1.y;
                }
                else
                {
                    delta = Coord(0, 255, 0);
                    dist = 256 - command.m_coord1.y;
                }
            }
            else if (command.m_coord1.z != 0)
            {
                if (command.m_coord1.z <= 15)
                {
                    delta = Coord(0, 0, 1);
                    dist = command.m_coord1.z;
                }
                else
                {
                    delta = Coord(0, 0, 255);
                    dist = 256 - command.m_coord1.z;
                }
            }
            newEnergy += 2 * dist;
            Coord coord = bot.m_pos;
            volatileCoords.push_back(coord);
            for (uint32_t i = 0; i < dist; i++)
            {
                coord += delta;
                if (m_model.get(coord.x, coord.y, coord.z))
                {
                    if (pMsg) *pMsg = "SMove: voxel is full";
                    return false;
                }
                volatileCoords.push_back(coord);
            }
            newBots.push_back(bot);
            auto& newBot = newBots.back();
            newBot.m_pos = coord;
            break;
        }
        case Action::LMove:
        {
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "LMove: bend out of range";
                return false;
            }
            newX += (int8_t)command.m_coord2.x;
            newY += (int8_t)command.m_coord2.y;
            newZ += (int8_t)command.m_coord2.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "LMove: destination out of range";
                return false;
            }
            Coord coord = bot.m_pos;
            volatileCoords.push_back(coord);
            for (uint32_t leg = 0; leg < 2; leg++)
            {
                Coord delta;
                uint32_t dist = 0;
                Coord commandCoord =
                    leg == 0 ? command.m_coord1 : command.m_coord2;
                if (commandCoord.x != 0)
                {
                    if (commandCoord.x <= 15)
                    {
                        delta = Coord(1, 0, 0);
                        dist = commandCoord.x;
                    }
                    else
                    {
                        delta = Coord(255, 0, 0);
                        dist = 256 - commandCoord.x;
                    }
                }
                else if (commandCoord.y != 0)
                {
                    if (commandCoord.y <= 15)
                    {
                        delta = Coord(0, 1, 0);
                        dist = commandCoord.y;
                    }
                    else
                    {
                        delta = Coord(0, 255, 0);
                        dist = 256 - commandCoord.y;
                    }
                }
                else if (commandCoord.z != 0)
                {
                    if (commandCoord.z <= 15)
                    {
                        delta = Coord(0, 0, 1);
                        dist = commandCoord.z;
                    }
                    else
                    {
                        delta = Coord(0, 0, 255);
                        dist = 256 - commandCoord.z;
                    }
                }
                newEnergy += 2 * dist;
                for (uint32_t i = 0; i < dist; i++)
                {
                    coord += delta;
                    if (m_model.get(coord.x, coord.y, coord.z))
                    {
                        if (pMsg) *pMsg = "LMove: voxel is full";
                        return false;
                    }
                    volatileCoords.push_back(coord);
                }
            }
            newEnergy += 2 * 2;
            newBots.push_back(bot);
            auto& newBot = newBots.back();
            newBot.m_pos = coord;
            break;
        }
        case Action::FusionP:
        {
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "FusionP: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;

            size_t iRec = 0;
            while (iRec < fusionRecs.size())
            {
                if (fusionRecs[iRec].m_coordP == bot.m_pos &&
                    fusionRecs[iRec].m_coordS == coord)
                {
                    break;
                }
                iRec++;
            }
            if (iRec == fusionRecs.size())
            {
                fusionRecs.emplace_back();
                auto& rec = fusionRecs.back();
                rec.m_coordP = bot.m_pos;
                rec.m_coordS = coord;
            }
            fusionRecs[iRec].m_iBotP = iBot;

            volatileCoords.push_back(bot.m_pos);
            volatileCoords.push_back(coord);
            newEnergy -= 24;
            break;
        }
        case Action::FusionS:
        {
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "FusionS: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;

            size_t iRec = 0;
            while (iRec < fusionRecs.size())
            {
                if (fusionRecs[iRec].m_coordP == coord &&
                    fusionRecs[iRec].m_coordS == bot.m_pos)
                {
                    break;
                }
                iRec++;
            }
            if (iRec == fusionRecs.size())
            {
                fusionRecs.emplace_back();
                auto& rec = fusionRecs.back();
                rec.m_coordP = coord;
                rec.m_coordS = bot.m_pos;
            }
            fusionRecs[iRec].m_iBotS = iBot;

            break;
        }
        case Action::Fission:
        {
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "Fission: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;
            if (m_model.get(coord.x, coord.y, coord.z))
            {
                if (pMsg) *pMsg = "Fission: voxel is full";
                return false;
            }
            uint32_t m = command.m_m;
            if ((uint32_t)bot.m_seeds.size() < m + 1)
            {
                if (pMsg) *pMsg = "Fission: too few seeds";
                return false;
            }
            volatileCoords.push_back(bot.m_pos);
            volatileCoords.push_back(coord);
            newBots.push_back(bot);
            auto& newBot = newBots.back();
            newBot.m_seeds.erase(newBot.m_seeds.begin(),
                                 newBot.m_seeds.begin() + (m + 1));
            newBots.push_back(bot);
            auto& newBot2 = newBots.back();
            newBot2.m_bid = bot.m_seeds[0];
            newBot2.m_pos = coord;
            newBot2.m_seeds.erase(newBot2.m_seeds.begin() + (m + 1),
                                  newBot2.m_seeds.end());
            newBot2.m_seeds.erase(newBot2.m_seeds.begin());
            newEnergy += 24;
            break;
        }
        case Action::Fill:
        {
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "Fill: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;
            volatileCoords.push_back(bot.m_pos);
            volatileCoords.push_back(coord);
            fills.emplace_back(coord, coord);
            newBots.push_back(bot);
            break;
        }
        case Action::Void:
        {
            if (!m_full)
            {
                if (pMsg) *pMsg = "Void not allowed in lighting round";
                return false;
            }
            int32_t newX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t newY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t newZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)newX >= res ||
                (uint32_t)newY >= res ||
                (uint32_t)newZ >= res)
            {
                if (pMsg) *pMsg = "Void: position out of range";
                return false;
            }
            Coord coord = bot.m_pos + command.m_coord1;
            volatileCoords.push_back(bot.m_pos);
            volatileCoords.push_back(coord);
            voids.emplace_back(coord, coord);
            newBots.push_back(bot);
            break;
        }
        case Action::GFill:
        {
            if (!m_full)
            {
                if (pMsg) *pMsg = "GFill not allowed in lighting round";
                return false;
            }
            int32_t ndX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t ndY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t ndZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)ndX >= res ||
                (uint32_t)ndY >= res ||
                (uint32_t)ndZ >= res)
            {
                if (pMsg) *pMsg = "GFill: nd position out of range";
                return false;
            }
            int32_t fdX = ndX + (int8_t)command.m_coord2.x;
            int32_t fdY = ndY + (int8_t)command.m_coord2.y;
            int32_t fdZ = ndZ + (int8_t)command.m_coord2.z;
            if ((uint32_t)fdX >= res ||
                (uint32_t)fdY >= res ||
                (uint32_t)fdZ >= res)
            {
                if (pMsg) *pMsg = "GFill: fd position out of range";
                return false;
            }
            Coord coordLow = Coord(std::min(ndX, fdX),
                                   std::min(ndY, fdY),
                                   std::min(ndZ, fdZ));
            Coord coordHigh = Coord(std::max(ndX, fdX),
                                    std::max(ndY, fdY),
                                    std::max(ndZ, fdZ));

            size_t iRec = 0;
            while (iRec < groupRecs.size())
            {
                if (groupRecs[iRec].m_coordLow == coordLow &&
                    groupRecs[iRec].m_coordHigh == coordHigh &&
                    groupRecs[iRec].m_action == action)
                {
                    break;
                }
                iRec++;
            }
            if (iRec == groupRecs.size())
            {
                groupRecs.emplace_back();
                auto& rec = groupRecs.back();
                rec.m_coordLow = coordLow;
                rec.m_coordHigh = coordHigh;
                rec.m_action = action;
            }
            groupRecs[iRec].m_iBots[ndX > fdX][ndY > fdY][ndZ > fdZ] = iBot;

            volatileCoords.push_back(bot.m_pos);
            newBots.push_back(bot);
            break;
        }
        case Action::GVoid:
        {
            if (!m_full)
            {
                if (pMsg) *pMsg = "GVoid not allowed in lighting round";
                return false;
            }
            int32_t ndX = (int32_t)bot.m_pos.x + (int8_t)command.m_coord1.x;
            int32_t ndY = (int32_t)bot.m_pos.y + (int8_t)command.m_coord1.y;
            int32_t ndZ = (int32_t)bot.m_pos.z + (int8_t)command.m_coord1.z;
            if ((uint32_t)ndX >= res ||
                (uint32_t)ndY >= res ||
                (uint32_t)ndZ >= res)
            {
                if (pMsg) *pMsg = "GVoid: nd position out of range";
                return false;
            }
            int32_t fdX = ndX + (int8_t)command.m_coord2.x;
            int32_t fdY = ndY + (int8_t)command.m_coord2.y;
            int32_t fdZ = ndZ + (int8_t)command.m_coord2.z;
            if ((uint32_t)fdX >= res ||
                (uint32_t)fdY >= res ||
                (uint32_t)fdZ >= res)
            {
                if (pMsg) *pMsg = "GVoid: fd position out of range";
                return false;
            }
            Coord coordLow = Coord(std::min(ndX, fdX),
                                   std::min(ndY, fdY),
                                   std::min(ndZ, fdZ));
            Coord coordHigh = Coord(std::max(ndX, fdX),
                                    std::max(ndY, fdY),
                                    std::max(ndZ, fdZ));

            size_t iRec = 0;
            while (iRec < groupRecs.size())
            {
                if (groupRecs[iRec].m_coordLow == coordLow &&
                    groupRecs[iRec].m_coordHigh == coordHigh &&
                    groupRecs[iRec].m_action == action)
                {
                    break;
                }
                iRec++;
            }
            if (iRec == groupRecs.size())
            {
                groupRecs.emplace_back();
                auto& rec = groupRecs.back();
                rec.m_coordLow = coordLow;
                rec.m_coordHigh = coordHigh;
                rec.m_action = action;
            }
            groupRecs[iRec].m_iBots[ndX > fdX][ndY > fdY][ndZ > fdZ] = iBot;

            volatileCoords.push_back(bot.m_pos);
            newBots.push_back(bot);
            break;
        }
        }
    }

    for (auto& rec : fusionRecs)
    {
        if (rec.m_iBotP == (size_t)-1 ||
            rec.m_iBotS == (size_t)-1)
        {
            if (pMsg) *pMsg = "Fusion mismatch";
            return false;
        }
        auto& botP = m_bots[rec.m_iBotP];
        auto& botS = m_bots[rec.m_iBotS];
        newBots.push_back(botP);
        auto& newBot = newBots.back();
        newBot.m_seeds.push_back(botS.m_bid);
        newBot.m_seeds.insert(newBot.m_seeds.end(),
                              botS.m_seeds.begin(),
                              botS.m_seeds.end());
        std::sort(newBot.m_seeds.begin(), newBot.m_seeds.end());
    }

    for (auto& rec : groupRecs)
    {
        bool diffX = rec.m_coordLow.x != rec.m_coordHigh.x;
        bool diffY = rec.m_coordLow.y != rec.m_coordHigh.y;
        bool diffZ = rec.m_coordLow.z != rec.m_coordHigh.z;
        for (uint32_t x = 0; x < 2; x++)
        {
            for (uint32_t y = 0; y < 2; y++)
            {
                for (uint32_t z = 0; z < 2; z++)
                {
                    if ((!x || diffX) &&
                        (!y || diffY) &&
                        (!z || diffZ) &&
                        rec.m_iBots[x][y][z] == (size_t)-1)
                    {
                        if (pMsg)
                        {
                            *pMsg =
                                rec.m_action == Action::GFill ?
                                "Group fill mismatch" :
                                "Group void mismatch";
                        }
                        return false;
                    }
                }
            }
        }
        /*
        if (rec.m_iBots[0][0][0] == (size_t)-1 ||
            rec.m_iBots[0][0][1] == (size_t)-1 ||
            rec.m_iBots[0][1][0] == (size_t)-1 ||
            rec.m_iBots[0][1][1] == (size_t)-1 ||
            rec.m_iBots[1][0][0] == (size_t)-1 ||
            rec.m_iBots[1][0][1] == (size_t)-1 ||
            rec.m_iBots[1][1][0] == (size_t)-1 ||
            rec.m_iBots[1][1][1] == (size_t)-1)
        {
            if (pMsg)
            {
                *pMsg =
                    rec.m_action == Action::GFill ?
                    "Group fill mismatch" :
                    "Group void mismatch";
            }
            return false;
        }
        */
        for (uint8_t x = rec.m_coordLow.x; x <= rec.m_coordHigh.x; x++)
        {
            for (uint8_t y = rec.m_coordLow.y; y <= rec.m_coordHigh.y; y++)
            {
                for (uint8_t z = rec.m_coordLow.z; z <= rec.m_coordHigh.z; z++)
                {
                    volatileCoords.push_back(Coord(x, y, z));
                }
            }
        }
        if (rec.m_action == Action::GFill)
        {
            fills.emplace_back(rec.m_coordLow, rec.m_coordHigh);
        }
        else
        {
            voids.emplace_back(rec.m_coordLow, rec.m_coordHigh);
        }
    }

    std::sort(volatileCoords.begin(), volatileCoords.end());
    auto adjIt = std::adjacent_find(volatileCoords.begin(), volatileCoords.end());
    if (adjIt != volatileCoords.end())
    {
        printf("Interfering coordinate: %s\n", adjIt->toString().c_str());
        if (pMsg) *pMsg = "Coordinates interfere";
        return false;
    }

    if (m_harmonics || newHarmonics || !voids.empty())
    {
        for (auto& fill : fills)
        {
            for (uint8_t x = fill.first.x; x <= fill.second.x; x++)
            {
                for (uint8_t y = fill.first.y; y <= fill.second.y; y++)
                {
                    for (uint8_t z = fill.first.z; z <= fill.second.z; z++)
                    {
                        if (m_model.get(x, y, z))
                        {
                            newEnergy += 6;
                        }
                        else
                        {
                            m_model.set(x, y, z, 1);
                            newEnergy += 12;
                        }
                    }
                }
            }
        }

        for (auto& void_ : voids)
        {
            for (uint8_t x = void_.first.x; x <= void_.second.x; x++)
            {
                for (uint8_t y = void_.first.y; y <= void_.second.y; y++)
                {
                    for (uint8_t z = void_.first.z; z <= void_.second.z; z++)
                    {
                        if (m_model.get(x, y, z))
                        {
                            m_model.set(x, y, z, 0);
                            newEnergy -= 12;
                        }
                        else
                        {
                            newEnergy += 3;
                        }
                    }
                }
            }
        }
    }
    else
    {
        for (auto& fill : fills)
        {
            newEnergy +=
                6 *
                (fill.second.x - fill.first.x + 1) *
                (fill.second.y - fill.first.y + 1) *
                (fill.second.z - fill.first.z + 1);
        }

        bool didSomething = true;
        bool skippedSomething = true;
        while (didSomething && skippedSomething)
        {
            didSomething = false;
            skippedSomething = false;
            for (auto& fill : fills)
            {
                for (uint8_t x = fill.first.x; x <= fill.second.x; x++)
                {
                    for (uint8_t y = fill.first.y; y <= fill.second.y; y++)
                    {
                        for (uint8_t z = fill.first.z; z <= fill.second.z; z++)
                        {
                            if (!m_model.get(x, y, z))
                            {
                                uint8_t ok =
                                    (y == 0) ||
                                    (y > 0 &&
                                     m_model.get(x, y - 1, z)) ||
                                    ((uint32_t)y + 1 < res &&
                                     m_model.get(x, y + 1, z)) ||
                                    (x > 0 &&
                                     m_model.get(x - 1, y, z)) ||
                                    ((uint32_t)x + 1 < res &&
                                     m_model.get(x + 1, y, z)) ||
                                    (z > 0 &&
                                     m_model.get(x, y, z - 1)) ||
                                    ((uint32_t)z + 1 < res &&
                                     m_model.get(x, y, z + 1));
                                if (ok)
                                {
                                    m_model.set(x, y, z, 1);
                                    newEnergy += 6;
                                    didSomething = true;
                                }
                                else
                                {
                                    skippedSomething = true;
                                }
                            }
                        }
                    }
                }
            }
        }
        if (skippedSomething)
        {
            if (pMsg) *pMsg = "Not grounded";
            return false;
        }
    }

    //if (!fills.empty() || !voids.empty())
    if (!newHarmonics && (m_harmonics || !voids.empty()))
    {
        Model tmpModel(m_model);

        vector<Coord> q;
        for (uint32_t x = 0; x < res; x++)
        {
            for (uint32_t z = 0; z < res; z++)
            {
                if (tmpModel.get(x, 0, z))
                {
                    tmpModel.set(x, 0, z, 0);
                    q.push_back(Coord(x, 0, z));
                }
            }
        }

        vector<Coord> q2;
        while (!q.empty())
        {
            for (auto& coord : q)
            {
                uint32_t x = coord.x;
                uint32_t y = coord.y;
                uint32_t z = coord.z;
                if (x > 0 && tmpModel.get(x - 1, y, z))
                {
                    tmpModel.set(x - 1, y, z, 0);
                    q2.push_back(Coord(x - 1, y, z));
                }
                if (x + 1 < res && tmpModel.get(x + 1, y, z))
                {
                    tmpModel.set(x + 1, y, z, 0);
                    q2.push_back(Coord(x + 1, y, z));
                }
                if (y > 0 && tmpModel.get(x, y - 1, z))
                {
                    tmpModel.set(x, y - 1, z, 0);
                    q2.push_back(Coord(x, y - 1, z));
                }
                if (y + 1 < res && tmpModel.get(x, y + 1, z))
                {
                    tmpModel.set(x, y + 1, z, 0);
                    q2.push_back(Coord(x, y + 1, z));
                }
                if (z > 0 && tmpModel.get(x, y, z - 1))
                {
                    tmpModel.set(x, y, z - 1, 0);
                    q2.push_back(Coord(x, y, z - 1));
                }
                if (z + 1 < res && tmpModel.get(x, y, z + 1))
                {
                    tmpModel.set(x, y, z + 1, 0);
                    q2.push_back(Coord(x, y, z + 1));
                }
            }
            q.clear();
            q.swap(q2);
        }

        for (uint32_t x = 0; x < res; x++)
        {
            for (uint32_t y = 0; y < res; y++)
            {
                for (uint32_t z = 0; z < res; z++)
                {
                    if (tmpModel.get(x, y, z))
                    {
                        if (pMsg) *pMsg = "Not grounded";
                        return false;
                    }
                }
            }
        }
    }

    std::sort(newBots.begin(), newBots.end(),
              [](const Bot& a, const Bot& b){return a.m_bid < b.m_bid;});

    m_bots.swap(newBots);
    m_energy = newEnergy;
    m_harmonics = newHarmonics;

    pos += numBots;
    return true;
}
