#include "Model.hpp"
#include "Cleanup.hpp"
#include "Command.hpp"

using std::string;
using std::vector;
using std::unique_ptr;

bool loadTraceData(const string& fileName,
                   vector<uint8_t>* pData,
                   string* pMsg)
{
    FILE* f = fopen(fileName.c_str(), "rb");
    if (!f)
    {
        if (pMsg) *pMsg = "Error opening trace file";
        return false;
    }
    Cleanup fileCleanup([&](){fclose(f);});

    fseek(f, 0, SEEK_END);
    size_t size = (size_t)ftell(f);
    fseek(f, 0, SEEK_SET);

    auto& data = *pData;
    data.resize(size);
    if (fread(data.data(), 1, size, f) != size)
    {
        if (pMsg) *pMsg = "Error reading trace data";
        return false;
    }
    fileCleanup.cleanup();

    return true;
}

bool parseTrace(bool full,
                const vector<uint8_t>& data,
                vector<Command>* pTrace,
                string* pMsg)
{
    auto& trace = *pTrace;
    size_t size = data.size();
    size_t pos = 0;
    while (pos < size)
    {
        uint8_t b = data[pos++];

        // Halt
        if (b == 0b11111111)
        {
            trace.push_back(Command(Action::Halt));
            continue;
        }

        // Wait
        if (b == 0b11111110)
        {
            trace.push_back(Command(Action::Wait));
            continue;
        }

        // Flip
        if (b == 0b11111101)
        {
            trace.push_back(Command(Action::Flip));
            continue;
        }

        // SMove
        if ((b & 0b1111) == 0b0100)
        {
            if (pos >= size)
            {
                if (pMsg) *pMsg = "Unexpected partial SMove command";
                return false;
            }
            uint8_t b2 = data[pos++];

            uint8_t a = (b >> 4) & 0b11;
            uint8_t i = b2 & 0b11111;
            if (i > 30)
            {
                if (pMsg) *pMsg = "Invalid SMove i value";
                return false;
            }
            if (a == 0b01)
            {
                trace.push_back(Command(Action::SMove, Coord(i - 15, 0, 0)));
            }
            else if (a == 0b10)
            {
                trace.push_back(Command(Action::SMove, Coord(0, i - 15, 0)));
            }
            else if (a == 0b11)
            {
                trace.push_back(Command(Action::SMove, Coord(0, 0, i - 15)));
            }
            else
            {
                if (pMsg) *pMsg = "Invalid SMove a value";
                return false;
            }
            continue;
        }

        // LMove
        if ((b & 0b1111) == 0b1100)
        {
            if (pos >= size)
            {
                if (pMsg) *pMsg = "Unexpected partial LMove command";
                return false;
            }
            uint8_t b2 = data[pos++];

            uint8_t a1 = (b >> 4) & 0b11;
            uint8_t a2 = (b >> 6) & 0b11;
            uint8_t i1 = b2 & 0b1111;
            uint8_t i2 = (b2 >> 4) & 0b1111;
            if (i1 > 10)
            {
                if (pMsg) *pMsg = "Invalid LMove i1 value";
                return false;
            }
            if (i2 > 10)
            {
                if (pMsg) *pMsg = "Invalid LMove i2 value";
                return false;
            }
            Coord coord1;
            if (a1 == 0b01)
            {
                coord1 = Coord(i1 - 5, 0, 0);
            }
            else if (a1 == 0b10)
            {
                coord1 = Coord(0, i1 - 5, 0);
            }
            else if (a1 == 0b11)
            {
                coord1 = Coord(0, 0, i1 - 5);
            }
            else
            {
                if (pMsg) *pMsg = "Invalid LMove a1 value";
                return false;
            }
            Coord coord2;
            if (a2 == 0b01)
            {
                coord2 = Coord(i2 - 5, 0, 0);
            }
            else if (a2 == 0b10)
            {
                coord2 = Coord(0, i2 - 5, 0);
            }
            else if (a2 == 0b11)
            {
                coord2 = Coord(0, 0, i2 - 5);
            }
            else
            {
                if (pMsg) *pMsg = "Invalid LMove a2 value";
                return false;
            }
            trace.push_back(Command(Action::LMove, coord1, coord2));
            continue;
        }

        // FusionP
        if ((b & 0b111) == 0b111)
        {
            uint8_t nd = (b >> 3) & 0b11111;
            if (nd > 26)
            {
                if (pMsg) *pMsg = "Invalid FusionP nd value";
                return false;
            }

            uint8_t z = nd % 3;
            uint8_t y = (nd / 3) % 3;
            uint8_t x = nd / 9;
            Coord coord(x - 1, y - 1, z - 1);
            trace.push_back(Command(Action::FusionP, coord));
            continue;
        }

        // FusionS
        if ((b & 0b111) == 0b110)
        {
            uint8_t nd = (b >> 3) & 0b11111;
            if (nd > 26)
            {
                if (pMsg) *pMsg = "Invalid FusionS nd value";
                return false;
            }

            uint8_t z = nd % 3;
            uint8_t y = (nd / 3) % 3;
            uint8_t x = nd / 9;
            Coord coord(x - 1, y - 1, z - 1);
            trace.push_back(Command(Action::FusionS, coord));
            continue;
        }

        // Fission
        if ((b & 0b111) == 0b101)
        {
            if (pos >= size)
            {
                if (pMsg) *pMsg = "Unexpected partial LMove command";
                return false;
            }
            uint8_t b2 = data[pos++];

            uint8_t nd = (b >> 3) & 0b11111;
            if (nd > 26)
            {
                if (pMsg) *pMsg = "Invalid Fission nd value";
                return false;
            }
            uint8_t m = b2;

            uint8_t z = nd % 3;
            uint8_t y = (nd / 3) % 3;
            uint8_t x = nd / 9;
            Coord coord(x - 1, y - 1, z - 1);
            trace.push_back(Command(Action::Fission, coord, m));
            continue;
        }

        // Fill
        if ((b & 0b111) == 0b011)
        {
            uint8_t nd = (b >> 3) & 0b11111;
            if (nd > 26)
            {
                if (pMsg) *pMsg = "Invalid Fill nd value";
                return false;
            }

            uint8_t z = nd % 3;
            uint8_t y = (nd / 3) % 3;
            uint8_t x = nd / 9;
            Coord coord(x - 1, y - 1, z - 1);
            trace.push_back(Command(Action::Fill, coord));
            continue;
        }

        // Void
        if ((b & 0b111) == 0b010)
        {
            if (!full)
            {
                if (pMsg) *pMsg = "Void not allowed in lighting round";
                return false;
            }

            uint8_t nd = (b >> 3) & 0b11111;
            if (nd > 26)
            {
                if (pMsg) *pMsg = "Invalid Void nd value";
                return false;
            }

            uint8_t z = nd % 3;
            uint8_t y = (nd / 3) % 3;
            uint8_t x = nd / 9;
            Coord coord(x - 1, y - 1, z - 1);
            trace.push_back(Command(Action::Void, coord));
            continue;
        }

        // GFill
        if ((b & 0b111) == 0b001)
        {
            if (!full)
            {
                if (pMsg) *pMsg = "GFill not allowed in lighting round";
                return false;
            }

            if (pos + 3 > size)
            {
                if (pMsg) *pMsg = "Unexpected partial GFill command";
                return false;
            }
            uint8_t fdx = data[pos++];
            uint8_t fdy = data[pos++];
            uint8_t fdz = data[pos++];

            uint8_t nd = (b >> 3) & 0b11111;
            if (nd > 26)
            {
                if (pMsg) *pMsg = "Invalid GFill nd value";
                return false;
            }
            if (fdx > 60 || fdy > 60 || fdz > 60 ||
                (fdx == 30 && fdy == 30 && fdz == 30))
            {
                if (pMsg) *pMsg = "Invalid GFill fd value";
                return false;
            }

            uint8_t z = nd % 3;
            uint8_t y = (nd / 3) % 3;
            uint8_t x = nd / 9;
            Coord coord1(x - 1, y - 1, z - 1);
            Coord coord2(fdx - 30, fdy - 30, fdz - 30);
            trace.push_back(Command(Action::GFill, coord1, coord2));
            continue;
        }

        // GVoid
        if ((b & 0b111) == 0b000)
        {
            if (!full)
            {
                if (pMsg) *pMsg = "GVoid not allowed in lighting round";
                return false;
            }

            if (pos + 3 > size)
            {
                if (pMsg) *pMsg = "Unexpected partial GVoid command";
                return false;
            }
            uint8_t fdx = data[pos++];
            uint8_t fdy = data[pos++];
            uint8_t fdz = data[pos++];

            uint8_t nd = (b >> 3) & 0b11111;
            if (nd > 26)
            {
                if (pMsg) *pMsg = "Invalid GVoid nd value";
                return false;
            }
            if (fdx > 60 || fdy > 60 || fdz > 60 ||
                (fdx == 30 && fdy == 30 && fdz == 30))
            {
                if (pMsg) *pMsg = "Invalid GVoid fd value";
                return false;
            }

            uint8_t z = nd % 3;
            uint8_t y = (nd / 3) % 3;
            uint8_t x = nd / 9;
            Coord coord1(x - 1, y - 1, z - 1);
            Coord coord2(fdx - 30, fdy - 30, fdz - 30);
            trace.push_back(Command(Action::GVoid, coord1, coord2));
            continue;
        }
    }

    return true;
}

bool loadTrace(bool full,
               const string& fileName,
               vector<Command>* pTrace,
               string* pMsg)
{
    vector<uint8_t> data;
    if (!loadTraceData(fileName, &data, pMsg))
    {
        return false;
    }

    if (!parseTrace(full, data, pTrace, pMsg))
    {
        return false;
    }

    return true;
}

bool formatTrace(bool full,
                 const vector<Command>& trace,
                 vector<uint8_t>* pData,
                 string* pMsg)
{
    auto& data = *pData;

    for (const Command& command : trace)
    {
        Action action = command.m_action;

        switch (action)
        {
        case Action::Halt:
        {
            data.push_back(0b11111111);
            break;
        }
        case Action::Wait:
        {
            data.push_back(0b11111110);
            break;
        }
        case Action::Flip:
        {
            data.push_back(0b11111101);
            break;
        }
        case Action::SMove:
        {
            Coord coord = command.m_coord1;
            uint8_t x = coord.x + 15;
            uint8_t y = coord.y + 15;
            uint8_t z = coord.z + 15;
            if (x != 15 && x <= 30 && y == 15 && z == 15)
            {
                data.push_back(0b00010100);
                data.push_back(x);
            }
            else if (y != 15 && y <= 30 && x == 15 && z == 15)
            {
                data.push_back(0b00100100);
                data.push_back(y);
            }
            else if (z != 15 && z <= 30 && x == 15 && y == 15)
            {
                data.push_back(0b00110100);
                data.push_back(z);
            }
            else
            {
                if (pMsg) *pMsg = "Invalid SMove coord";
                return false;
            }
            break;
        }
        case Action::LMove:
        {
            uint8_t a1 = 0;
            uint8_t i1 = 0;
            {
                auto& a = a1;
                auto& i = i1;
                Coord coord = command.m_coord1;
                uint8_t x = coord.x + 5;
                uint8_t y = coord.y + 5;
                uint8_t z = coord.z + 5;
                if (x != 5 && x <= 10 && y == 5 && z == 5)
                {
                    a = 0b01;
                    i = x;
                }
                else if (y != 5 && y <= 10 && x == 5 && z == 5)
                {
                    a = 0b10;
                    i = y;
                }
                else if (z != 5 && z <= 10 && x == 5 && y == 5)
                {
                    a = 0b11;
                    i = z;
                }
                else
                {
                    if (pMsg) *pMsg = "Invalid LMove coord1";
                    return false;
                }
            }

            uint8_t a2 = 0;
            uint8_t i2 = 0;
            {
                auto& a = a2;
                auto& i = i2;
                Coord coord = command.m_coord2;
                uint8_t x = coord.x + 5;
                uint8_t y = coord.y + 5;
                uint8_t z = coord.z + 5;
                if (x != 5 && x <= 10 && y == 5 && z == 5)
                {
                    a = 0b01;
                    i = x;
                }
                else if (y != 5 && y <= 10 && x == 5 && z == 5)
                {
                    a = 0b10;
                    i = y;
                }
                else if (z != 5 && z <= 10 && x == 5 && y == 5)
                {
                    a = 0b11;
                    i = z;
                }
                else
                {
                    if (pMsg) *pMsg = "Invalid LMove coord2";
                    return false;
                }
            }

            data.push_back(0b00001100 | (a1 << 4) | (a2 << 6));
            data.push_back(i1 | (i2 << 4));

            break;
        }
        case Action::FusionP:
        {
            Coord coord = command.m_coord1;
            uint32_t count = (coord.x != 0) + (coord.y != 0) + (coord.z != 0);
            uint8_t x = coord.x + 1;
            uint8_t y = coord.y + 1;
            uint8_t z = coord.z + 1;
            if ((count != 1 && count != 2) ||
                x > 2 || y > 2 || z > 2)
            {
                if (pMsg) *pMsg = "Invalid FusionP coord";
                return false;
            }
            uint8_t nd = x * 9 + y * 3 + z;
            data.push_back(0b00000111 | (nd << 3));
            break;
        }
        case Action::FusionS:
        {
            Coord coord = command.m_coord1;
            uint32_t count = (coord.x != 0) + (coord.y != 0) + (coord.z != 0);
            uint8_t x = coord.x + 1;
            uint8_t y = coord.y + 1;
            uint8_t z = coord.z + 1;
            if ((count != 1 && count != 2) ||
                x > 2 || y > 2 || z > 2)
            {
                if (pMsg) *pMsg = "Invalid FusionS coord";
                return false;
            }
            uint8_t nd = x * 9 + y * 3 + z;
            data.push_back(0b00000110 | (nd << 3));
            break;
        }
        case Action::Fission:
        {
            Coord coord = command.m_coord1;
            uint32_t count = (coord.x != 0) + (coord.y != 0) + (coord.z != 0);
            uint8_t x = coord.x + 1;
            uint8_t y = coord.y + 1;
            uint8_t z = coord.z + 1;
            if ((count != 1 && count != 2) ||
                x > 2 || y > 2 || z > 2)
            {
                if (pMsg) *pMsg = "Invalid Fission coord";
                return false;
            }
            uint8_t nd = x * 9 + y * 3 + z;
            data.push_back(0b00000101 | (nd << 3));
            data.push_back(command.m_m);
            break;
        }
        case Action::Fill:
        {
            Coord coord = command.m_coord1;
            uint32_t count = (coord.x != 0) + (coord.y != 0) + (coord.z != 0);
            uint8_t x = coord.x + 1;
            uint8_t y = coord.y + 1;
            uint8_t z = coord.z + 1;
            if ((count != 1 && count != 2) ||
                x > 2 || y > 2 || z > 2)
            {
                if (pMsg) *pMsg = "Invalid Fill coord";
                return false;
            }
            uint8_t nd = x * 9 + y * 3 + z;
            data.push_back(0b00000011 | (nd << 3));
            break;
        }
        case Action::Void:
        {
            if (!full)
            {
                if (pMsg) *pMsg = "Void not allowed in lighting round";
                return false;
            }
            Coord coord = command.m_coord1;
            uint32_t count = (coord.x != 0) + (coord.y != 0) + (coord.z != 0);
            uint8_t x = coord.x + 1;
            uint8_t y = coord.y + 1;
            uint8_t z = coord.z + 1;
            if ((count != 1 && count != 2) ||
                x > 2 || y > 2 || z > 2)
            {
                if (pMsg) *pMsg = "Invalid Void coord";
                return false;
            }
            uint8_t nd = x * 9 + y * 3 + z;
            data.push_back(0b00000010 | (nd << 3));
            break;
        }
        case Action::GFill:
        {
            if (!full)
            {
                if (pMsg) *pMsg = "GFill not allowed in lighting round";
                return false;
            }
            Coord coord1 = command.m_coord1;
            uint32_t count = (coord1.x != 0) + (coord1.y != 0) + (coord1.z != 0);
            uint8_t x = coord1.x + 1;
            uint8_t y = coord1.y + 1;
            uint8_t z = coord1.z + 1;
            if ((count != 1 && count != 2) ||
                x > 2 || y > 2 || z > 2)
            {
                if (pMsg) *pMsg = "Invalid GFill coord1";
                return false;
            }
            Coord coord2 = command.m_coord2;
            uint8_t fdx = coord2.x + 30;
            uint8_t fdy = coord2.y + 30;
            uint8_t fdz = coord2.z + 30;
            if (fdx > 60 || fdy > 60 || fdz > 60 ||
                (fdx == 30 && fdy == 30 && fdz == 30))
            {
                if (pMsg) *pMsg = "Invalid GFill coord2";
                return false;
            }
            uint8_t nd = x * 9 + y * 3 + z;
            data.push_back(0b00000001 | (nd << 3));
            data.push_back(fdx);
            data.push_back(fdy);
            data.push_back(fdz);
            break;
        }
        case Action::GVoid:
        {
            if (!full)
            {
                if (pMsg) *pMsg = "GVoid not allowed in lighting round";
                return false;
            }
            Coord coord1 = command.m_coord1;
            uint32_t count = (coord1.x != 0) + (coord1.y != 0) + (coord1.z != 0);
            uint8_t x = coord1.x + 1;
            uint8_t y = coord1.y + 1;
            uint8_t z = coord1.z + 1;
            if ((count != 1 && count != 2) ||
                x > 2 || y > 2 || z > 2)
            {
                if (pMsg) *pMsg = "Invalid GVoid coord1";
                return false;
            }
            Coord coord2 = command.m_coord2;
            uint8_t fdx = coord2.x + 30;
            uint8_t fdy = coord2.y + 30;
            uint8_t fdz = coord2.z + 30;
            if (fdx > 60 || fdy > 60 || fdz > 60 ||
                (fdx == 30 && fdy == 30 && fdz == 30))
            {
                if (pMsg) *pMsg = "Invalid GVoid coord2";
                return false;
            }
            uint8_t nd = x * 9 + y * 3 + z;
            data.push_back(0b00000000 | (nd << 3));
            data.push_back(fdx);
            data.push_back(fdy);
            data.push_back(fdz);
            break;
        }
        default:
            if (pMsg) *pMsg = "Invalid action";
            return false;
        }
    }

    return true;
}

bool saveTraceData(const vector<uint8_t>& data,
                   const string& fileName,
                   string* pMsg)
{
    FILE* f = fopen(fileName.c_str(), "wb");
    if (!f)
    {
        if (pMsg) *pMsg = "Error opening trace file";
        return false;
    }
    Cleanup fileCleanup([&](){fclose(f);});

    if (!data.empty())
    {
        if (fwrite(data.data(), 1, data.size(), f) != data.size())
        {
            if (pMsg) *pMsg = "Error writing trace file";
            return false;
        }
    }

    fileCleanup.cleanup();

    return true;
}

bool saveTrace(bool full,
               const vector<Command>& trace,
               const string& fileName,
               string* pMsg)
{
    vector<uint8_t> data;
    if (!formatTrace(full, trace, &data, pMsg))
    {
        return false;
    }

    if (!saveTraceData(data, fileName, pMsg))
    {
        return false;
    }

    return true;
}
