#include "Common.hpp"
#include "StringUtils.hpp"
#include "Model.hpp"
#include "Command.hpp"
#include "Trace.hpp"
#include "Sim.hpp"
#include "Check.hpp"
#include "Solve.hpp"
#include "Batch.hpp"
#include "Mapping.hpp"
#include "Optimize.hpp"

using std::string;
using std::vector;

#define OPTIMIZE_BEST 0
#define OPTIMIZE_SOLVED 1

void usage()
{
    printf("Usage: solve <batch> [<start> [<end>]]\n");
    printf("Batches:\n");
    printf("  la    Lightning assemble                    (range: 1-186)\n");
    printf("  fa    Full assemble                         (range: 1-186)\n");
    printf("  fd    Full disassemble                      (range: 1-186)\n");
    printf("  fr    Full reassemble                       (range: 1-115)\n");
    printf("  frd   Full reassemble: disassemble portion  (range: 1-115)\n");
    printf("  fra   Full reassemble: assemble portion     (range: 1-115)\n");
}

int main(int argc, char *argv[])
{
    if (argc < 2 || argc > 4)
    {
        usage();
        return 1;
    }

    string batchStr = argv[1];
    uint32_t start = 1;
    uint32_t end = 186;
    bool full = true;
    Batch batch = Batch::FA;
    if (!parseBatch(batchStr, &batch))
    {
        usage();
    }

    if (batch == Batch::LA)
    {
        full = false;
    }
    else if (batch == Batch::FA)
    {
    }
    else if (batch == Batch::FD)
    {
    }
    else if (batch == Batch::FR ||
             batch == Batch::FRD ||
             batch == Batch::FRA)
    {
        end = 115;
    }
    else
    {
        usage();
        return 1;
    }

    if (argc >= 3)
    {
        uint32_t tmpStart = (uint32_t)atoi(argv[2]);
        if (tmpStart < start || tmpStart > end)
        {
            usage();
            return 1;
        }
        start = tmpStart;
    }

    if (argc >= 4)
    {
        uint32_t tmpEnd = (uint32_t)atoi(argv[3]);
        if (tmpEnd < start || tmpEnd > end)
        {
            usage();
            return 1;
        }
        end = tmpEnd;
    }

    string msg;

    vector<Mapping> frdMappings = getMappings(Batch::FRD);
    vector<Mapping> fraMappings = getMappings(Batch::FRA);

    for (uint32_t i = start; i <= end; i++)
    {
        printf("\n");

        Model sourceModel;
        Model targetModel;
        string defaultTraceName;
        string bestTraceName;

        if (batch == Batch::LA)
        {
            string targetModelName = strprintf("../../../data/problemsL/LA%03" PRIu32 "_tgt.mdl", i);
            if (!targetModel.load(targetModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            sourceModel.resize(targetModel.getRes());

            defaultTraceName = strprintf("../../../data/dfltTracesL/LA%03" PRIu32 ".nbt", i);
            bestTraceName = strprintf("../../../data/bestL/LA%03" PRIu32 ".nbt", i);
        }

        if (batch == Batch::FA)
        {
            string targetModelName = strprintf("../../../data/problemsF/FA%03" PRIu32 "_tgt.mdl", i);
            if (!targetModel.load(targetModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            sourceModel.resize(targetModel.getRes());

            defaultTraceName = strprintf("../../../data/dfltTracesF/FA%03" PRIu32 ".nbt", i);
            bestTraceName = strprintf("../../../data/bestF/FA%03" PRIu32 ".nbt", i);
        }

        if (batch == Batch::FD)
        {
            string sourceModelName = strprintf("../../../data/problemsF/FD%03" PRIu32 "_src.mdl", i);
            if (!sourceModel.load(sourceModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            targetModel.resize(sourceModel.getRes());

            defaultTraceName = strprintf("../../../data/dfltTracesF/FD%03" PRIu32 ".nbt", i);
            bestTraceName = strprintf("../../../data/bestF/FD%03" PRIu32 ".nbt", i);
        }

        if (batch == Batch::FR)
        {
            string sourceModelName = strprintf("../../../data/problemsF/FR%03" PRIu32 "_src.mdl", i);
            if (!sourceModel.load(sourceModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            string targetModelName = strprintf("../../../data/problemsF/FR%03" PRIu32 "_tgt.mdl", i);
            if (!targetModel.load(targetModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            if (sourceModel.getRes() != targetModel.getRes())
            {
                printf("%03" PRIu32": Resolution mismatch\n", i);
                continue;
            }

            defaultTraceName = strprintf("../../../data/dfltTracesF/FR%03" PRIu32 ".nbt", i);
            bestTraceName = strprintf("../../../data/bestF/FR%03" PRIu32 ".nbt", i);
        }

        if (batch == Batch::FRD)
        {
            auto& mapping = frdMappings[i - 1];
            if (!(mapping.m_batch == batch && mapping.m_idx == i))
            {
                printf("%03" PRIu32": Skipping\n", i);
                continue;
            }

            string sourceModelName = strprintf("../../../data/problemsF/FR%03" PRIu32 "_src.mdl", i);
            if (!sourceModel.load(sourceModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            targetModel.resize(sourceModel.getRes());

            defaultTraceName = "";
            bestTraceName = strprintf("../../../data/portionsF/FRD%03" PRIu32 ".nbt", i);
        }

        if (batch == Batch::FRA)
        {
            auto& mapping = fraMappings[i - 1];
            if (!(mapping.m_batch == batch && mapping.m_idx == i))
            {
                printf("%03" PRIu32": Skipping\n", i);
                continue;
            }

            string targetModelName = strprintf("../../../data/problemsF/FR%03" PRIu32 "_tgt.mdl", i);
            if (!targetModel.load(targetModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            sourceModel.resize(targetModel.getRes());

            defaultTraceName = "";
            bestTraceName = strprintf("../../../data/portionsF/FRA%03" PRIu32 ".nbt", i);
        }

        // Default trace

        if (batch != Batch::FRD &&
            batch != Batch::FRA)
        {
            vector<Command> defaultTrace;
            if (!loadTrace(full, defaultTraceName, &defaultTrace, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            printf("%03" PRIu32": Default trace energy:", i);
            fflush(stdout);

            uint64_t defaultEnergy = 0;
#if 1
            if (!checkTrace(full, sourceModel, targetModel, defaultTrace, &defaultEnergy, &msg))
            {
                printf("\n");
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }
#endif
            printf("  %16" PRIu64 "\n", defaultEnergy);
        }


        // Best trace

        vector<Command> bestTrace;
        bool gotBestTrace = false;
        if (!loadTrace(full, bestTraceName, &bestTrace, &msg))
        {
            printf("%03" PRIu32": %s\n", i, msg.c_str());
            if (batch != Batch::FRD &&
                batch != Batch::FRA)
            {
                continue;
            }
        }
        else
        {
            gotBestTrace = true;
        }

        uint64_t bestEnergy = 0;
        if (gotBestTrace)
        {
            printf("%03" PRIu32": Best trace energy:", i);
            fflush(stdout);

#if 1
            if (!checkTrace(full, sourceModel, targetModel, bestTrace, &bestEnergy, &msg))
            {
                printf("\n");
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }
#endif
            printf("     %16" PRIu64 "\n", bestEnergy);

#if OPTIMIZE_BEST
            if (batch != Batch::FR)
            {
                printf("%03" PRIu32": Optimizing harmonics...", i);
                fflush(stdout);

                vector<Command> optimizedBestTrace;
                if (!optimizeHarmonics(full, sourceModel, bestTrace, &optimizedBestTrace, &msg))
                {
                    printf("\r\x1b[K");
                    printf("%03" PRIu32": %s\n", i, msg.c_str());
                    continue;
                }
                printf("\r\x1b[K");

                printf("%03" PRIu32": Optimized best:", i);
                fflush(stdout);

                uint64_t optimizedBestEnergy = 0;
                if (!checkTrace(full, sourceModel, targetModel, optimizedBestTrace, &optimizedBestEnergy, &msg))
                {
                    printf("\n");
                    printf("%03" PRIu32": %s\n", i, msg.c_str());
                    continue;
                }
                bool newBestOptimized = optimizedBestEnergy < bestEnergy;
                printf("        %16" PRIu64 "%s\n", optimizedBestEnergy,
                       newBestOptimized ? "  NEW BEST!" : "");

#if 1
                if (newBestOptimized && batch != Batch::LA)
                {
                    if (!saveTrace(full, optimizedBestTrace, bestTraceName, &msg))
                    {
                        printf("%03" PRIu32": %s\n", i, msg.c_str());
                        continue;
                    }
                }
#endif
            }
#endif
        }

#if OPTIMIZE_BEST
        continue;
#endif


        // Solve

        vector<Command> trace;

        if (batch == Batch::FR)
        {
            auto& frdMapping = frdMappings[i - 1];
            string frdTraceName =
                frdMapping.m_batch == Batch::FD ?
                strprintf("../../../data/bestF/FD%03" PRIu32 ".nbt", frdMapping.m_idx) :
                frdMapping.m_batch == Batch::FRD ?
                strprintf("../../../data/portionsF/FRD%03" PRIu32 ".nbt", frdMapping.m_idx) :
                "";
            vector<Command> frdTrace;
            if (!loadTrace(full, frdTraceName, &frdTrace, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }
            if (frdTrace.empty())
            {
                printf("%03" PRIu32": Trace empty\n", i);
                continue;
            }

            auto& fraMapping = fraMappings[i - 1];
            string fraTraceName =
                fraMapping.m_batch == Batch::FA ?
                strprintf("../../../data/bestF/FA%03" PRIu32 ".nbt", fraMapping.m_idx) :
                fraMapping.m_batch == Batch::FRA ?
                strprintf("../../../data/portionsF/FRA%03" PRIu32 ".nbt", fraMapping.m_idx) :
                "";
            vector<Command> fraTrace;
            if (!loadTrace(full, fraTraceName, &fraTrace, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }
            if (fraTrace.empty())
            {
                printf("%03" PRIu32": Trace empty\n", i);
                continue;
            }

            trace.resize((frdTrace.size() - 1) + fraTrace.size());
            std::copy(frdTrace.begin(),
                      frdTrace.begin() + (frdTrace.size() - 1),
                      trace.begin());
            std::copy(fraTrace.begin(),
                      fraTrace.end(),
                      trace.begin() + (frdTrace.size() - 1));
        }
        else
        {
            printf("%03" PRIu32": Solving...", i);
            fflush(stdout);

            solve(full, sourceModel, targetModel, &trace);
            printf("\r\x1b[K");

        }

#if 0
        printf("%03" PRIu32": Solution trace size:   %16" PRIuZ "\n",
               i, trace.size());
#endif

#if OPTIMIZE_SOLVED
        if (batch != Batch::FR)
        {
            bool usesHarmonics = false;
            for (auto& command : trace)
            {
                if (command.m_action == Action::Flip)
                {
                    usesHarmonics = true;
                    break;
                }
            }

            if (usesHarmonics)
            {
                printf("%03" PRIu32": Optimizing harmonics...", i);
                fflush(stdout);

                vector<Command> optimizedTrace;
                if (!optimizeHarmonics(full, sourceModel, trace, &optimizedTrace, &msg))
                {
                    printf("\r\x1b[K");
                    printf("%03" PRIu32": %s\n", i, msg.c_str());
                    continue;
                }
                printf("\r\x1b[K");

                trace.swap(optimizedTrace);
            }
        }
#endif

        printf("%03" PRIu32": Checking solution...", i);
        fflush(stdout);

        uint64_t energy = 0;
        if (!checkTrace(full, sourceModel, targetModel, trace, &energy, &msg))
        {
            printf("\r\x1b[K");
            printf("%03" PRIu32": %s\n", i, msg.c_str());
            continue;
        }
        printf("\r\x1b[K");

        bool newBest = !gotBestTrace || energy < bestEnergy;
        bool same = gotBestTrace && energy == bestEnergy;
        printf("%03" PRIu32": Solution trace energy: %16" PRIu64 "%s\n",
               i,
               energy,
               newBest ? "  NEW BEST!" :
               same ? "  (same)" : "");

        // Save best

#if 1
        if (newBest && batch != Batch::LA)
        {
            if (!saveTrace(full, trace, bestTraceName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }
        }
#endif
    }

    return 0;
}
